/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"

INT PX(local_size_dft_3d_internal)(
    const INT *n, const PX(profile) *prf,
    const PX(truncation) *trc, const PX(permutation) *perm,
    INT howmany, const INT *block_col, const INT *block_row,
    MPI_Comm comm_col, MPI_Comm comm_row,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start
    )
{ /* get local data size of given profile */
  int num_comm;
  INT alloc_local_transp[3], alloc_local_fft[3], alloc_local_max, local_n[3], n_trunc[3];
  const INT *block;
  PX(permutation) perm_out;
  PX(truncation) trc_out;
  MPI_Comm comm;
  
  PX(copy_permutation)(perm, &perm_out);
  PX(init_truncation)(trc->ni, trc->no, &trc_out);
  for(int p=0; p<3; p++){
    PX(decompose_fft_trunc_3d)(
        n, &trc_out, &perm_out, block_col, block_row, comm_col, comm_row,
        local_ni, local_n, local_no);

    alloc_local_fft[p] = PX(local_size_fft_1d)(
        local_n, howmany, prf->fft_flag[p], &trc_out, &perm_out);
    
    num_comm = PX(remap_num_comm)(prf->first_dims[p], prf->remap_flag[p], &perm_out);
    block = (num_comm == 0) ? block_col : block_row;
    comm  = (num_comm == 0) ? comm_col  : comm_row;
    PX(get_current_global_size)(&trc_out, n_trunc);

    alloc_local_transp[p] = PX(local_size_transpose_2d)(
        n_trunc, local_no, howmany, block, prf->first_dims[p], comm, prf->remap_flag[p], &perm_out);
  }

  /* get initial and final data offsets */
  PX(decompose_3d)(trc->ni, perm, block_col, block_row, comm_col, comm_row,
      local_ni, local_i_start);
  PX(decompose_3d)(trc->no, &perm_out, block_col, block_row, comm_col, comm_row,
      local_no, local_o_start);

  /* return maximum needed memory */
  alloc_local_max = 0;
  for(int t=0; t<3; t++){
    alloc_local_max = MAX(alloc_local_max, alloc_local_transp[t]);
    alloc_local_max = MAX(alloc_local_max, alloc_local_fft[t]);
  }

  return alloc_local_max;
}


PX(plan) PX(plan_dft_internal)(
    int rank, const INT *n, const PX(profile) *prf,
    const PX(truncation) *trc, const PX(permutation) *perm,
    INT howmany, const INT *block_col, const INT *block_row,
    C *data_in, C *data_out,
    MPI_Comm comm_col, MPI_Comm comm_row, MPI_Comm comm_msg,
    int sign, unsigned fftw_flags
    )
{ /* !!! with flag FFTW_MEASURE planning will overwrite users data !!! */
  int num_comm;
  INT prod_last_dims=1;
  INT local_ni[3], local_n[3], local_no[3], n_trunc[3];
  const INT *block;
  PX(truncation) trc_out;
  PX(plan) ths;
  MPI_Comm comm;
  
  if(rank < 3)
    return NULL;
  
  for(int t=3; t<rank; t++)
    prod_last_dims *= n[t];

  ths = (plan_s*) PX(malloc)(sizeof(plan_s), comm_msg);
  ths->n = (INT*) PX(malloc)(sizeof(INT)*rank, comm_msg);
  ths->fftw_flags = fftw_flags;
  ths->sign = sign;
  ths->rank = rank;
  for(int t=0; t<rank; t++)
    ths->n[t] = n[t];
  for(int t=0; t<3; t++){
    ths->ni[t] = trc->ni[t];
    ths->no[t] = trc->no[t];
  }
  MPI_Comm_size(comm_col, &ths->np[0]);
  MPI_Comm_size(comm_row, &ths->np[1]);
  ths->np[2] = 1;
  
  PX(copy_permutation)(perm, &ths->perm_in);
  PX(copy_profile)(prf, &ths->prf);
  
  PX(copy_permutation)(perm, &ths->perm_out);
  PX(init_truncation)(trc->ni, trc->no, &trc_out);
  
  /* split planing of 1dFFTs and remaps to avoid long planing
  times caused by forgotten wisdom in front of remap planing */
  for(int p=0; p<3; p++){
    /* get data layout */
    PX(decompose_fft_trunc_3d)(
        n, &trc_out, &ths->perm_out, block_col, block_row, comm_col, comm_row,
        local_ni, local_n, local_no);

    if(p==0)
      ths->fft_last_dims = PX(plan_fft_for_last_dims)(
          rank, n, howmany, local_ni, data_in, data_in,
          sign, fftw_flags);
    
    /* plan 1D-FFT */
    ths->local_fft[p] = PX(plan_dft_trunc_1d)(
        local_ni, local_n, local_no, howmany*prod_last_dims, data_in, data_out,
        sign, prf->fft_flag[p], fftw_flags,
        &trc_out, &ths->perm_out);

    /* simulate remap */
    num_comm = PX(remap_num_comm)(prf->first_dims[p], prf->remap_flag[p], &ths->perm_out);
    block = (num_comm == 0) ? block_col : block_row;
    comm  = (num_comm == 0) ? comm_col  : comm_row;

    PX(permute_transpose_2d)(prf->first_dims[p], prf->remap_flag[p], &ths->perm_out);
  }
  
  /* reinitialize data distribution and permutation */
  PX(copy_permutation)(perm, &ths->perm_out);
  PX(init_truncation)(trc->ni, trc->no, &trc_out);
  for(int p=0; p<3; p++){
    /* get data layout */
    PX(decompose_fft_trunc_3d)(
        n, &trc_out, &ths->perm_out, block_col, block_row, comm_col, comm_row,
        local_ni, local_n, local_no);

    /* simulate 1D-FFT */
    PX(permute_fft_1d)(prf->fft_flag[p], &ths->perm_out);
    PX(mark_performed_fft)(&ths->perm_out, &trc_out);
    
    /* plan remap */
    X(forget_wisdom)();
    num_comm = PX(remap_num_comm)(prf->first_dims[p], prf->remap_flag[p], &ths->perm_out);
    block = (num_comm == 0) ? block_col : block_row;
    comm  = (num_comm == 0) ? comm_col  : comm_row;
    PX(get_current_global_size)(&trc_out, n_trunc);
    
    ths->global_remap[p] =  PX(plan_transpose_2d)(
        n_trunc, local_no, howmany*prod_last_dims, block, data_out, data_in, prf->first_dims[p], comm,
        fftw_flags| prf->remap_flag[p], &ths->perm_out);

    /* FIXME: ??? possible FFTW BUG ????:
     * something goes wrong in FFTWs wisdom management for parallel transposition,
     * sometimes next planing freezes, delete wisdom to avoid this */
    X(forget_wisdom)();
  }
  
  /* set all times to zero */
  PX(reset_timer)(ths);
  
  if( PX(is_illegal_plan)(ths) ){
      PX(destroy_plan)(ths);
      ths = NULL;
  }

  return ths;
}

void PX(permute_fft_3d)(
    const PX(profile) *prf, PX(permutation) *perm
    )
{
  for(int p=0; p<3; p++){
    PX(permute_fft_1d)(prf->fft_flag[p], perm);
    PX(permute_transpose_2d)(prf->first_dims[p], prf->remap_flag[p], perm);
  }
}


