/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

// #include "api.h"
// #include "dft.h"
// #include "rdft.h"

#include <stddef.h> /* for ptrdiff_t */

#include <pfft.h>
#include <ipfft.h>
#include "fortran-x.h"

/* if F77_FUNC is not defined, then we don't know how to mangle identifiers
   for the Fortran linker, and we must omit the f77 API. */
#if defined(F77_FUNC) || defined(WINDOWS_F77_MANGLING)

/*-----------------------------------------------------------------------*/
/* some internal functions used by the f77 api */

/* in fortran, the natural array ordering is column-major, which
   corresponds to reversing the dimensions relative to C's row-major */
static void revert_int(int rnk, const int *n, int *n_rev)
{
  for (int i = 0; i < 3; i++)
    n_rev[2-i] = n[i];
  for (int i = 3; i < rnk; i++)
    n_rev[i] = n[i];
}

static void revert_INT(int rnk, const ptrdiff_t *n, ptrdiff_t *n_rev)
{
  for (int i = 0; i < 3; i++)
    n_rev[2-i] = n[i];
  for (int i = 3; i < rnk; i++)
    n_rev[i] = n[i];
}

static void revert_and_sub_ones_INT(int rnk, const ptrdiff_t *n, ptrdiff_t *n_rev)
{
  revert_INT(rnk, n, n_rev);
  for (int i = 0; i < rnk; i++)
    n_rev[i]--;
}

static void revert_and_add_ones_INT(int rnk, const ptrdiff_t *n, ptrdiff_t *n_rev)
{
  revert_INT(rnk, n, n_rev);
  for (int i = 0; i < rnk; i++)
    n_rev[i]++;
}

static ptrdiff_t *malloc_INT(int rnk)
{
  return (ptrdiff_t *) X(malloc)(sizeof(ptrdiff_t) * rnk);
}

static ptrdiff_t *malloc_and_revert_INT(int rnk, const ptrdiff_t *n)
{
  ptrdiff_t *n_rev;
  
  if(n == NULL)
    return NULL;
  
  n_rev = malloc_INT(rnk);
  revert_INT(rnk, n, n_rev);
  
  return n_rev;
}

static void revert_perm(
    const PX(permutation) *perm,
    PX(permutation) *perm_rev
    )
{
  revert_int(3, perm->np, perm_rev->np);
  revert_int(3, perm->n, perm_rev->n);
}

static ptrdiff_t *malloc_and_revert_blocks(
    int rnk, const ptrdiff_t *blk
    )
{
  if(blk[0] == FPFFT_DEFAULT_BLOCKS)
    return PFFT_DEFAULT_BLOCKS;
  else
    return malloc_and_revert_INT(rnk, blk);
}

static void free_blocks(
    ptrdiff_t **blk
    )
{
  if( *blk != PFFT_DEFAULT_BLOCKS )
    PX(free)(*blk);
}

static ptrdiff_t *malloc_and_revert_gcells(
    int rnk, const ptrdiff_t *gcells
    )
{
  if(gcells[0] == FPFFT_NO_GCELLS)
    return PFFT_NO_GCELLS;
  else
    return malloc_and_revert_INT(rnk, gcells);
}

static void free_gcells(
    ptrdiff_t **gcells
    )
{
  if( *gcells != PFFT_NO_GCELLS )
    PX(free)(*gcells);
}

/*-----------------------------------------------------------------------*/

#  define F77(a, A) F77px(px77(a), PX77(A))

#  ifndef WINDOWS_F77_MANGLING

#    if defined(F77_FUNC)
#      define F77px(a, A) F77_FUNC(a, A)
#      include "fortran-funcs.h"
#    endif

/* If identifiers with underscores are mangled differently than those
   without underscores, then we include *both* mangling versions.  The
   reason is that the only Fortran compiler that does such differing
   mangling is currently g77 (which adds an extra underscore to names
   with underscores), whereas other compilers running on the same
   machine are likely to use non-underscored mangling.  (I'm sick
   of users complaining that FFTW works with g77 but not with e.g.
   pgf77 or ifc on the same machine.)  Note that all FFTW identifiers
   contain underscores, and configure picks g77 by default. */
#    if defined(F77_FUNC_) && !defined(F77_FUNC_EQUIV)
#      undef F77px
#      define F77px(a, A) F77_FUNC_(a, A)
#      include "fortran-funcs.h"
#    endif

#  else /* WINDOWS_F77_MANGLING */

/* Various mangling conventions common (?) under Windows. */

/* g77 */
#    define WINDOWS_F77_FUNC(a, A) a ## __
#    define F77px(a, A) WINDOWS_F77_FUNC(a, A)
#    include "fortran-funcs.h"

/* Intel, etc. */
#    undef WINDOWS_F77_FUNC
#    define WINDOWS_F77_FUNC(a, A) a ## _
#    include "fortran-funcs.h"

/* Digital/Compaq/HP Visual Fortran, Intel Fortran.  stdcall attribute
   is apparently required to adjust for calling conventions (callee
   pops stack in stdcall).  See also:
       http://msdn.microsoft.com/library/en-us/vccore98/html/_core_mixed.2d.language_programming.3a_.overview.asp
*/
#    undef WINDOWS_F77_FUNC
#    if defined(__GNUC__)
#      define WINDOWS_F77_FUNC(a, A) __attribute__((stdcall)) A
#    elif defined(_MSC_VER) || defined(_ICC) || defined(_STDCALL_SUPPORTED)
#      define WINDOWS_F77_FUNC(a, A) __stdcall A
#    else
#      define WINDOWS_F77_FUNC(a, A) A /* oh well */
#    endif
#    include "fortran-funcs.h"

#  endif /* WINDOWS_F77_MANGLING */

#endif /* F77_FUNC || WINDOWS_F77_MANGLING */
