/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"

static void exchange_gcells_along_one_dim(
    PX(gcplan) ths, int dim);
static void reduce_gcells_along_one_dim(
    PX(gcplan) ths, int dim);
static void exchange_gcells_above_along_one_dim(
    PX(gcplan) ths, int dim);
static void exchange_gcells_below_along_one_dim(
    PX(gcplan) ths, int dim);
static void reduce_gcells_above_along_one_dim(
    PX(gcplan) ths, int dim);
static void reduce_gcells_below_along_one_dim(
    PX(gcplan) ths, int dim);
static void sendrecv_gcells_along_one_dim(
    PX(gcplan) ths, int dim, int dir,
    INT numSend, INT sendOffset, INT numRecv, INT recvOffset);
static void addsendrecv_gcells_along_one_dim(
    PX(gcplan) ths, int dim, int dir,
    INT numSend, INT sendOffset, INT numRecv, INT recvOffset);
static void isend_slices(
    R *data, const INT *localArraySize3d, int dim, int dir, INT tupleSize,
    INT numSend, INT sendOffset, MPI_Comm commCart1d, MPI_Request *request);
static void irecv_slices(
    R *data, const INT *localArraySize3d, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d, MPI_Request *request);
static void addrecv_slices(
    R *data, const INT *localArraySize3d, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d);
static void create_mpi_datatype_slices(
    const INT *localArraySize3d, INT tupleSize,
    int dim, int numSlices, int offset,
    MPI_Datatype *newtype);
static void add_buffer_to_slices(
    R *data, R *buffer,
    const INT *localArraySize3d, INT tupleSize,
    int dim, int numSlices, int offset);
static void irecv_slices(
    R *data, const INT *localArraySize3d, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d, MPI_Request *request);


void PX(exchange_gcells_internal_sendrecv)(
    PX(gcplan) ths
    )
{
  for(int dim=0; dim<3; dim++)
    exchange_gcells_along_one_dim(ths, dim);
}


void PX(reduce_gcells_internal_sendrecv)(
    PX(gcplan) ths
    )
{
  for(int dim=0; dim<3; dim++)
    reduce_gcells_along_one_dim(ths, dim);

  PX(truncate_array_outside)(
      ths->ngc, ths->gc_below, ths->gc_above, ths->tuple,
      ths->data);
}


static void exchange_gcells_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  exchange_gcells_below_along_one_dim(ths, dim);
  exchange_gcells_above_along_one_dim(ths, dim);
}


static void reduce_gcells_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  reduce_gcells_below_along_one_dim(ths, dim);
  reduce_gcells_above_along_one_dim(ths, dim);
}


static void exchange_gcells_above_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  INT numSendLeftOver, numRecvLeftOver, sendOffset, recvOffset;
  INT numCurrentSend, numCurrentRecv, numSendAvail, numRecvAvail;
  INT localArrayStart, localArrayEnd;
  INT globalArraySize = ths->n[dim], blockSize = ths->blk[dim];

  localArrayStart  = localArrayEnd = ths->gc_below[dim];
  localArrayEnd   += ths->loc_n[dim];

  sendOffset = localArrayStart;
  recvOffset = localArrayEnd;
  numSendLeftOver = numRecvLeftOver = ths->gc_above[dim];
  for(int shift=0; (numSendLeftOver > 0) || (numRecvLeftOver > 0); shift++ ){
    numSendAvail = PX(local_block_size_shifted)(globalArraySize, blockSize, shift,   ths->comm_cart_1d[dim]);
    numRecvAvail = PX(local_block_size_shifted)(globalArraySize, blockSize, shift+1, ths->comm_cart_1d[dim]);

    numCurrentSend = MIN(numSendLeftOver, numSendAvail);
    numCurrentRecv = MIN(numRecvLeftOver, numRecvAvail);

    sendrecv_gcells_along_one_dim(
        ths, dim, -1, numCurrentSend, sendOffset, numCurrentRecv, recvOffset);

    sendOffset += numCurrentSend;
    recvOffset += numCurrentRecv;

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;
  }
}


static void exchange_gcells_below_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  INT numSendLeftOver, numRecvLeftOver, sendOffset, recvOffset;
  INT numCurrentSend, numCurrentRecv, numSendAvail, numRecvAvail;
  INT localArrayStart, localArrayEnd;
  INT globalArraySize = ths->n[dim], blockSize = ths->blk[dim];

  localArrayStart  = localArrayEnd = ths->gc_below[dim];
  localArrayEnd   += ths->loc_n[dim];

  sendOffset = localArrayEnd;
  recvOffset = localArrayStart;
  numSendLeftOver = numRecvLeftOver = ths->gc_below[dim];
  for(int shift=0; (numSendLeftOver > 0) || (numRecvLeftOver > 0); shift-- ){
    numSendAvail = PX(local_block_size_shifted)(globalArraySize, blockSize, shift,   ths->comm_cart_1d[dim]);
    numRecvAvail = PX(local_block_size_shifted)(globalArraySize, blockSize, shift-1, ths->comm_cart_1d[dim]);

    numCurrentSend = MIN(numSendLeftOver, numSendAvail);
    numCurrentRecv = MIN(numRecvLeftOver, numRecvAvail);

    sendOffset -= numCurrentSend;
    recvOffset -= numCurrentRecv;

    sendrecv_gcells_along_one_dim(
        ths, dim, +1, numCurrentSend, sendOffset, numCurrentRecv, recvOffset);

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;
  }
}


static void reduce_gcells_above_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  INT numSendLeftOver, numRecvLeftOver, sendOffset, recvOffset;
  INT numCurrentSend, numCurrentRecv, numSendAvail, numRecvAvail;
  INT localArrayStart, localArrayEnd, numSendPossible, numRecvPossible;
  INT globalArraySize = ths->n[dim], blockSize = ths->blk[dim];
  int numShifts;

  localArrayStart  = localArrayEnd = ths->gc_below[dim];
  localArrayEnd   += ths->loc_n[dim];


  numShifts = 0;
  numSendPossible = numRecvPossible = 0;
  numSendLeftOver = numRecvLeftOver = ths->gc_above[dim];
  for(int shift=0; (numSendLeftOver > 0) || (numRecvLeftOver > 0); shift++){
    numSendAvail = PX(local_block_size_shifted)(globalArraySize, blockSize, shift+1, ths->comm_cart_1d[dim]);
    numRecvAvail = PX(local_block_size_shifted)(globalArraySize, blockSize, shift,   ths->comm_cart_1d[dim]);

    numSendPossible += numSendAvail;
    numRecvPossible += numRecvAvail;

    numCurrentSend = MIN(numSendLeftOver, numSendAvail);
    numCurrentRecv = MIN(numRecvLeftOver, numRecvAvail);

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;

    numShifts++;
  }

  sendOffset = localArrayEnd + ths->gc_above[dim];
  recvOffset = localArrayStart + ths->gc_above[dim];
  numSendLeftOver = numRecvLeftOver = ths->gc_above[dim];
  for(int shift = numShifts-1; shift >= 0; shift-- ){
    numSendPossible -= PX(local_block_size_shifted)(globalArraySize, blockSize, shift+1, ths->comm_cart_1d[dim]);
    numRecvPossible -= PX(local_block_size_shifted)(globalArraySize, blockSize, shift,   ths->comm_cart_1d[dim]);

    numCurrentSend = numSendLeftOver - MIN(numSendLeftOver, numSendPossible);
    numCurrentRecv = numRecvLeftOver - MIN(numRecvLeftOver, numRecvPossible);

    sendOffset -= numCurrentSend;
    recvOffset -= numCurrentRecv;

    addsendrecv_gcells_along_one_dim(
        ths, dim, +1, numCurrentSend, sendOffset, numCurrentRecv, recvOffset);

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;

  }
}


static void reduce_gcells_below_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  INT numSendLeftOver, numRecvLeftOver, sendOffset, recvOffset;
  INT numCurrentSend, numCurrentRecv, numSendAvail, numRecvAvail;
  INT localArrayStart, localArrayEnd, numSendPossible, numRecvPossible;
  INT globalArraySize = ths->n[dim], blockSize = ths->blk[dim];
  int numShifts;

  localArrayStart  = localArrayEnd = ths->gc_below[dim];
  localArrayEnd   += ths->loc_n[dim];


  numShifts = 0;
  numSendPossible = numRecvPossible = 0;
  numSendLeftOver = numRecvLeftOver = ths->gc_below[dim];
  for(int shift=0; (numSendLeftOver > 0) || (numRecvLeftOver > 0); shift++){
    numSendAvail = PX(local_block_size_shifted)(globalArraySize, blockSize, shift,   ths->comm_cart_1d[dim]);
    numRecvAvail = PX(local_block_size_shifted)(globalArraySize, blockSize, shift+1, ths->comm_cart_1d[dim]);

    numSendPossible += numSendAvail;
    numRecvPossible += numRecvAvail;

    numCurrentSend = MIN(numSendLeftOver, numSendAvail);
    numCurrentRecv = MIN(numRecvLeftOver, numRecvAvail);

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;

    numShifts++;
  }

  sendOffset = 0;
  recvOffset = localArrayEnd - ths->gc_below[dim];
  numSendLeftOver = numRecvLeftOver = ths->gc_below[dim];
  for(int shift = numShifts-1; shift >= 0; shift-- ){
    numSendPossible -= PX(local_block_size_shifted)(globalArraySize, blockSize, shift,   ths->comm_cart_1d[dim]);
    numRecvPossible -= PX(local_block_size_shifted)(globalArraySize, blockSize, shift+1, ths->comm_cart_1d[dim]);

    numCurrentSend = numSendLeftOver - MIN(numSendLeftOver, numSendPossible);
    numCurrentRecv = numRecvLeftOver - MIN(numRecvLeftOver, numRecvPossible);

    addsendrecv_gcells_along_one_dim(
        ths, dim, -1, numCurrentSend, sendOffset, numCurrentRecv, recvOffset);

    sendOffset += numCurrentSend;
    recvOffset += numCurrentRecv;

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;
  }
}


static void sendrecv_gcells_along_one_dim(
    PX(gcplan) ths, int dim, int dir,
    INT numSend, INT sendOffset, INT numRecv, INT recvOffset
    )
{
  INT localArrayPlusGhostSize[3];
  MPI_Request mpi_req[2];

  for(int t=0; t<3; t++)
    localArrayPlusGhostSize[t] =
        ths->loc_n[t] + ths->gc_below[t] + ths->gc_above[t];

  irecv_slices(ths->data, localArrayPlusGhostSize, dim, dir,
      ths->tuple, numRecv, recvOffset, ths->comm_cart_1d[dim], &mpi_req[1]);
  isend_slices(ths->data, localArrayPlusGhostSize, dim, dir,
      ths->tuple, numSend, sendOffset, ths->comm_cart_1d[dim], &mpi_req[0]);
  MPI_Waitall(2, mpi_req, MPI_STATUSES_IGNORE);
}


static void addsendrecv_gcells_along_one_dim(
    PX(gcplan) ths, int dim, int dir,
    INT numSend, INT sendOffset, INT numRecv, INT recvOffset
    )
{
  INT localArrayPlusGhostSize[3];
  MPI_Request request;

  for(int t=0; t<3; t++)
    localArrayPlusGhostSize[t] =
        ths->loc_n[t] + ths->gc_below[t] + ths->gc_above[t];

  if( numSend > 0 )
    isend_slices(ths->data, localArrayPlusGhostSize, dim, dir,
        ths->tuple, numSend, sendOffset, ths->comm_cart_1d[dim], &request);
  if( numRecv > 0 )
    addrecv_slices(ths->data, localArrayPlusGhostSize, dim, dir,
        ths->tuple, numRecv, recvOffset, ths->comm_cart_1d[dim]);
  if( numSend > 0 )
    MPI_Wait(&request, MPI_STATUS_IGNORE);
}



static void isend_slices(
    R *data, const INT *localArraySize3d, int dim, int dir, INT tupleSize,
    INT numSend, INT sendOffset, MPI_Comm commCart1d, MPI_Request *request
    )
{
  int from, to, disp;
  MPI_Datatype sendtype;

  if(numSend == 0){
    *request = MPI_REQUEST_NULL;
    return;
  }

  create_mpi_datatype_slices(localArraySize3d, tupleSize, dim, numSend, sendOffset, &sendtype);
  MPI_Type_commit(&sendtype);

  disp = (dir<0) ? -1 : +1;
  MPI_Cart_shift(commCart1d, 0, disp, &from, &to);

  MPI_Isend(data, 1, sendtype, to, 7, commCart1d, request);

  MPI_Type_free(&sendtype);
}


static void irecv_slices(
    R *data, const INT *localArraySize3d, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d, MPI_Request *request
    )
{
  int from, to, disp;
  MPI_Datatype recvtype;

  if(numRecv == 0){
    *request = MPI_REQUEST_NULL;
    return;
  }

  create_mpi_datatype_slices(localArraySize3d, tupleSize, dim, numRecv, recvOffset, &recvtype);
  MPI_Type_commit(&recvtype);

  disp = (dir<0) ? -1 : +1;
  MPI_Cart_shift(commCart1d, 0, -disp, &to, &from);

  MPI_Irecv(data, 1, recvtype, from, 7, commCart1d, request);

  MPI_Type_free(&recvtype);
}



static void addrecv_slices(
    R *data, const INT *localArraySize3d, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d
    )
{
  int from, to, disp;
  INT numRecvTotal;
  R *buffer;

  numRecvTotal=tupleSize;
  for(int t=0; t<3; t++)
    numRecvTotal *= (dim==t) ? numRecv : localArraySize3d[t];

  disp = (dir<0) ? -1 : +1;
  MPI_Cart_shift(commCart1d, 0, -disp, &to, &from);

  buffer = (R*) fftw_malloc(sizeof(R)*numRecvTotal);
  MPI_Recv(buffer, numRecvTotal, PFFT_MPI_TYPE, from, 7, commCart1d, MPI_STATUS_IGNORE);

  add_buffer_to_slices(data, buffer, localArraySize3d, tupleSize, dim, numRecv, recvOffset);
  fftw_free(buffer);
}


static void create_mpi_datatype_slices(
    const INT *localArraySize3d, INT tupleSize,
    int dim, int numSlices, int offset,
    MPI_Datatype *newtype
    )
{
  const int d=4;
  int array_of_sizes[d], array_of_subsizes[d], array_of_starts[d];
  MPI_Datatype oldtype=PFFT_MPI_TYPE;

  for(int t=0; t<3; t++){
    array_of_sizes[t]    = (int) ( localArraySize3d[t] );
    array_of_subsizes[t] = (int) ( (t==dim) ? numSlices : localArraySize3d[t] );
    array_of_starts[t]   = (int) ( (t==dim) ? offset    : 0 );
  }
  array_of_sizes[3] = array_of_subsizes[3] = (int) (tupleSize);
  array_of_starts[3]= 0;

  /* FIXME: array_of_sizes, array_of_subsizes, array_of_starts are integer */
  /* TODO:  implement MPI_Type_create_subarray with INT similar MPI_Datatypes*/
  MPI_Type_create_subarray(d, array_of_sizes, array_of_subsizes, array_of_starts, MPI_ORDER_C, oldtype, newtype);
}


static void add_buffer_to_slices(
    R *data, R *buffer,
    const INT *localArraySize3d, INT tupleSize,
    int dim, int numSlices, int offset
    )
{
  const int d=4;
  INT n[d], nSub[d], nStart[d];
  INT k0, k1, k2, h, ind;

  for(int t=0; t<3; t++){
    n[t]      = localArraySize3d[t];
    nSub[t]   = (t==dim) ? numSlices : localArraySize3d[t];
    nStart[t] = (t==dim) ? offset    : 0;
  }

  ind = 0;
  for(k0 = nStart[0]; k0 < nStart[0] + nSub[0]; k0++){
    for(k1 = nStart[1]; k1 < nStart[1] + nSub[1]; k1++){
      for(k2 = nStart[2]; k2 < nStart[2] + nSub[2]; k2++){
        for(h = 0; h < tupleSize; h++){
          data[h+tupleSize*(k2+n[2]*(k1+n[1]*k0))] += buffer[ind++];
        }
      }
    }
  }
}

