#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <math.h>
#include <complex.h>
/* return value 0 on success */

#include <fftw3.h>
#include <fftw3-mpi.h>
#include <pfft.h>

#define MAX(a,b) (((a)>(b))?(a):(b))

static int allocate_fftw_complex(
    ptrdiff_t alloc_local, fftw_complex **data, char *name);

static void pfft_test_gcells(
    const ptrdiff_t *n, const int *np,
    const ptrdiff_t *gcells_below, const ptrdiff_t *gcells_above,
    MPI_Comm comm, unsigned fftw_flags, int verbose);
static void pfft_minimal_test_gcells(
    const ptrdiff_t *n, const int *np,
    const ptrdiff_t *gcells_below, const ptrdiff_t *gcells_above,
    MPI_Comm comm, unsigned fftw_flags);

static void init_parameters(
    int argc, char **argv,
    ptrdiff_t *n, int *np, ptrdiff_t *gcells_below, ptrdiff_t *gcells_above,
    int* planing, int* verbose
    )
{
  pfft_get_args(argc, argv, "-pfft_n", 3, PFFT_PTRDIFF_T, n);
  pfft_get_args(argc, argv, "-pfft_np", 2, PFFT_INT, np);
  pfft_get_args(argc, argv, "-pfft_gcells_below", 3, PFFT_PTRDIFF_T, gcells_below);
  pfft_get_args(argc, argv, "-pfft_gcells_above", 3, PFFT_PTRDIFF_T, gcells_above);
  pfft_get_args(argc, argv, "-pfft_planing", 1, PFFT_INT, planing);
  pfft_get_args(argc, argv, "-pfft_verbose", 1, PFFT_INT, verbose);
}


int main(int argc, char **argv){
  ptrdiff_t n[3], gcells_below[3], gcells_above[3];
  int np[2], myrank, size, planing, verbose;
  unsigned fftw_flags=0;
  
  MPI_Init(&argc, &argv);
  pfft_init();
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  
  /* default values */
  n[0] = n[1] = n[2] = 128; /*  n[0] = 3; n[1] = 5; n[2] = 7;*/
  np[0]=size; np[1]=1;
  verbose = 0;
  for(int t=0; t<3; t++){
    gcells_below[t] = 4;
    gcells_above[t] = 5;
  }
  
  /* this option causes a bug with np=[2,2] and n=[3,3,3]*/
//   fftw_set_timelimit(120.0);
  
  /* set values by commandline */
  init_parameters(argc, argv, n, np, gcells_below, gcells_above, &planing, &verbose);

  switch(planing){
    case 0: fftw_flags = FFTW_ESTIMATE; break;
    case 2: fftw_flags = FFTW_PATIENT; break;
    case 3: fftw_flags = FFTW_EXHAUSTIVE; break;
    default: fftw_flags = FFTW_MEASURE;
  }

  pfft_test_gcells(
      n, np, gcells_below, gcells_above, MPI_COMM_WORLD, fftw_flags, verbose);
  pfft_minimal_test_gcells(
      n, np, gcells_below, gcells_above, MPI_COMM_WORLD, fftw_flags);
  
  pfft_cleanup();
  MPI_Finalize();
  return 0;
}



static void pfft_test_gcells(
    const ptrdiff_t *n, const int *np,
    const ptrdiff_t *gcells_below, const ptrdiff_t *gcells_above,
    MPI_Comm comm, unsigned fftw_flags, int verbose
    )
{
  ptrdiff_t alloc_local, alloc_local_gc, local_ngc[3], local_gc_start[3];
  ptrdiff_t local_ni[3], local_i_start[3], local_no[3], local_o_start[3];
  fftw_complex *data;
  MPI_Comm comm_cart_2d;
  pfft_plan fftplan;
  pfft_gcplan gcplan;
  
  if( pfft_create_procmesh_2d(comm, np[0], np[1], &comm_cart_2d) )
    return;

  /* allocate mem for data */
  alloc_local = pfft_local_size_3d(n, comm_cart_2d, PFFT_TRANSPOSED,
      local_ni, local_i_start, local_no, local_o_start);
  alloc_local_gc = pfft_local_size_gcells_3d(
      local_no, local_o_start, alloc_local, gcells_below, gcells_above,
      local_ngc, local_gc_start);
  
  allocate_fftw_complex(alloc_local_gc, &data, "data");

  fftplan = pfft_plan_dft_3d(n, data, data, comm_cart_2d,
      FFTW_FORWARD, PFFT_TRANSPOSED| PFFT_FORWARD, fftw_flags);
  gcplan = pfft_plan_gcells_3d(n, gcells_below, gcells_above,
      data, comm_cart_2d, PFFT_TRANSPOSED| PFFT_GC_AFTER_FFT);
  
  pfft_init_input_3d(n, local_ni, local_i_start, comm_cart_2d, data);
  pfft_execute(fftplan);
  
  pfft_init_input_3d(n, local_no, local_o_start, comm_cart_2d, data);
  
  /* check gcell input */
  if(verbose)
    pfft_apr_complex_permuted_3d(data, local_no, local_o_start, 2, 1, 0, "gcell input", comm);
  
  /** add for gcell support */
  pfft_exchange(gcplan);
  
  /* check output */
  if(verbose)
    pfft_apr_complex_permuted_3d(data, local_ngc, local_gc_start, 2, 1, 0, "exchanged gcells", comm);
  
  pfft_reduce(gcplan);

  pfft_print_average_timer(comm, fftplan);
  pfft_print_average_gctimer(comm, gcplan);

  /* check input */
  if(verbose)
    pfft_apr_complex_permuted_3d(data, local_no, local_o_start, 2, 1, 0, "reduced gcells", comm);
  
  /** free mem and finalize */
  pfft_destroy_plan(fftplan);
  pfft_destroy_gcplan(gcplan);
  fftw_free(data);
  MPI_Comm_free(&comm_cart_2d);
}


static void pfft_minimal_test_gcells(
    const ptrdiff_t *n, const int *np,
    const ptrdiff_t *gcells_below, const ptrdiff_t *gcells_above,
    MPI_Comm comm, unsigned fftw_flags
    )
{
  ptrdiff_t alloc_local, alloc_local_gc, local_ngc[3], local_gc_start[3];
  ptrdiff_t local_ni[3], local_i_start[3], local_no[3], local_o_start[3];
  fftw_complex *data;
  MPI_Comm comm_cart_2d;
  pfft_plan plan_forw, plan_back;
  pfft_gcplan gcplan;

  if( pfft_create_procmesh_2d(comm, np[0], np[1], &comm_cart_2d) )
    return;

  alloc_local = pfft_local_size_3d(n, comm_cart_2d, PFFT_TRANSPOSED,
      local_ni, local_i_start, local_no, local_o_start);
  alloc_local_gc = pfft_local_size_gcells_3d(
      local_no, local_o_start, alloc_local, gcells_below, gcells_above,
      local_ngc, local_gc_start);
  
  /* allocate mem for data */
  allocate_fftw_complex(alloc_local_gc, &data, "data");

  plan_forw = pfft_plan_dft_3d(n, data, data, comm_cart_2d,
      FFTW_FORWARD,  PFFT_TRANSPOSED| PFFT_FORWARD, fftw_flags);
  plan_back = pfft_plan_dft_3d(n, data, data, comm_cart_2d,
      FFTW_BACKWARD, PFFT_TRANSPOSED| PFFT_BACKWARD, fftw_flags);
  gcplan = pfft_plan_gcells_3d(n, gcells_below, gcells_above,
      data, comm_cart_2d, PFFT_TRANSPOSED| PFFT_GC_AFTER_FFT);

  /* initialize data */
  pfft_init_input_3d(n, local_ni, local_i_start, comm_cart_2d, data);

  /* compute parallel forward FFT */
  pfft_execute(plan_forw);

  /* send ghostcells */
  pfft_exchange(gcplan);

  /* do some local calculations */

  /* sum up ghostcells */
  pfft_reduce(gcplan);

  /* compute parallel backward FFT */
  pfft_execute(plan_back);

  /** free mem and finalize */
  pfft_destroy_plan(plan_forw);
  pfft_destroy_plan(plan_back);
  pfft_destroy_gcplan(gcplan);
  fftw_free(data);
  MPI_Comm_free(&comm_cart_2d);
}


static int allocate_fftw_complex(
    ptrdiff_t alloc_local, fftw_complex **data, char *name
    )
{
  *data = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*alloc_local);
  if( *data == NULL )
    pfft_fprintf(MPI_COMM_WORLD, stderr, "!!! Error in PFFT: Not enough memory to allocate %s !!!\n", name);

  return data == NULL;
}
