/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"
#include <string.h>     /* strcmp */


void PX(get_timer)(
    PX(plan) ths, PX(timer) *timer
    )
{
  PX(copy_timer)(&ths->timer, timer);
}

void PX(convert_timer2vec)(
    const PX(timer) *timer,
    double *times
    )
{
  times[0] = (double) timer->iter;
  times[1] = timer->whole;
  for(int t=0; t<4; t++)
    times[t+2] = timer->fft[t];
  for(int t=0; t<3; t++)
    times[t+6] = timer->remap[t];
}


void PX(convert_vec2timer)(
    const double *times,
    PX(timer) *timer
    )
{
  timer->iter  = round(times[0]);
  timer->whole = times[1];
  for(int t=0; t<4; t++)
    timer->fft[t]   = times[t+2];
  for(int t=0; t<3; t++)
    timer->remap[t] = times[t+6];
}


void PX(average_timer)(
     PX(timer) *timer
     )
{
  double times[9];
  
  PX(convert_timer2vec)(timer, times);
  if(times[0] != 0){
    for(int t=1; t<9; t++)
      times[t] /= times[0];
    times[0] = 1;
  }
  PX(convert_vec2timer)(times, timer);
}

void PX(copy_timer)(
    const PX(timer) *orig,
    PX(timer) *copy
    )
{
  double times[9];
  
  PX(convert_timer2vec)(orig, times);
  PX(convert_vec2timer)(times, copy);
}

void PX(reduce_max_timer)(
    MPI_Comm comm, const PX(timer) *timer,
    PX(timer) *timer_max
    )
{
  double times[9], times_max[9];
  
  PX(convert_timer2vec)(timer, times);
  MPI_Reduce(times, times_max, 9, MPI_DOUBLE, MPI_MAX, 0, comm);
  PX(convert_vec2timer)(times_max, timer_max);
}

void PX(add_timers)(
    const PX(timer) *sum1, const PX(timer) *sum2,
    PX(timer) *res
    )
{
  double t1[9], t2[9], sum[9];
  
  PX(convert_timer2vec)(sum1, t1);
  PX(convert_timer2vec)(sum2, t2);
  for(int t=0; t<9; t++)
    sum[t] = t1[t] + t2[t];
  PX(convert_vec2timer)(sum, res);
}


static void reset_timer(
    PX(timer) *timer
    )
{
  double times[9];
  
  for(int t=0; t<9; t++)
    times[t] = 0;
  PX(convert_vec2timer)(times, timer);
}

void PX(reset_timer)(
    PX(plan) ths
    )
{
  reset_timer(&ths->timer);
}


static int file_exists(
    const char *restrict name
    )
{
  FILE *f;
  
  f=fopen(name, "r");
  if(f != NULL)
    fclose(f);
  
  return (f != NULL);
}

static FILE* open_or_create_file_to_append(
    MPI_Comm comm, const char *restrict name
    )
{
  FILE *f=fopen(name, "a+");
  if(f==NULL){
    PX(fprintf)(comm, stderr, "Error: Cannot open file %s.\n", name);
    exit(1);
  }
  return f;
}

static void write_info_header(
    MPI_Comm comm, FILE *restrict file
    )
{
  PX(fprintf)(comm, file, "\n%% n - FFT size\n");
  PX(fprintf)(comm, file, "%% np - process grid\n");
  PX(fprintf)(comm, file, "%% procs - number of processes\n");
  PX(fprintf)(comm, file, "%% pfft - PFFT runtime\n");
  PX(fprintf)(comm, file, "%% index(i) = log(procs(i)) + 1\n");
}

static int get_matlab_index(
    MPI_Comm comm
    )
{
  int size;
  MPI_Comm_size(comm, &size);
  return (int)round(log(size)/log(2))+1;
}

static void write_run_specific_infos(
    MPI_Comm comm, FILE *restrict file, PX(plan) ths
    )
{
  
  int size;
  int idx = get_matlab_index(comm);
  MPI_Comm_size(comm, &size);
  
  if(ths->fftw_flags & FFTW_ESTIMATE)
    PX(fprintf)(comm, file, "\n%% fftw_flags == FFTW_ESTIMATE\n");
  else if(ths->fftw_flags & FFTW_PATIENT)
    PX(fprintf)(comm, file, "\n%% fftw_flags == FFTW_PATIENT\n");
  else if(ths->fftw_flags & FFTW_EXHAUSTIVE)
    PX(fprintf)(comm, file, "\n%% fftw_flags == FFTW_EXHAUSTIVE\n");
  else
    PX(fprintf)(comm, file, "\n%% fftw_flags == FFTW_MEASURE\n");
  
  PX(fprintf)(comm, file, "index(%d) = %d;  ", idx, idx);
  PX(fprintf)(comm, file, "procs(%d) = %d;  ", idx, size);
  PX(fprintf)(comm, file, "np_pfft(%d, 1:3) = [%d %d %d];\n", idx, ths->np[0], ths->np[1], ths->np[2]);
  PX(fprintf)(comm, file, "n_pfft(%d, 1:%d) = [", idx, ths->rank);
  for(int t=0; t<ths->rank; t++)
    PX(fprintf)(comm, file, "%td ", ths->n[t]);
  PX(fprintf)(comm, file, "];  ");
  PX(fprintf)(comm, file, "ni_pfft(%d, 1:3) = [%td %td %td];  ", idx, ths->ni[0], ths->ni[1], ths->ni[2]);
  PX(fprintf)(comm, file, "no_pfft(%d, 1:3) = [%td %td %td];\n", idx, ths->no[0], ths->no[1], ths->no[2]);
}

static void fprint_average_timer_internal(
    MPI_Comm comm, FILE *restrict file, const char *prefix, PX(timer) *ths,
    unsigned flags
    )
{
  int idx = get_matlab_index(comm);
  PX(timer) mt;
  PX(reduce_max_timer)(comm, ths, &mt);
  PX(average_timer)(&mt);
  
  if(flags & PFFT_PRINT_TIMER_BASIC){
    PX(fprintf)(comm, file, "%s_fft_iter(%d)    = %d;  ", prefix, idx, ths->iter);
    PX(fprintf)(comm, file, "%s(%d)   = %.3e;\n", prefix, idx, mt.whole);
  } else if(flags & PFFT_PRINT_TIMER_ADV){
    PX(fprintf)(comm, file, "%s_fft1(%d, 1)   = %.3e;  ", prefix, idx, mt.fft[0]);
    PX(fprintf)(comm, file, "%s_remap1(%d, 1) = %.3e;\n", prefix, idx, mt.remap[0]);
    PX(fprintf)(comm, file, "%s_fft2(%d, 2)   = %.3e;  ", prefix, idx, mt.fft[1]);
    PX(fprintf)(comm, file, "%s_remap2(%d, 2) = %.3e;\n", prefix, idx, mt.remap[1]);
    PX(fprintf)(comm, file, "%s_fft3(%d, 3)   = %.3e;  ", prefix, idx, mt.fft[2]);
    PX(fprintf)(comm, file, "%s_remap3(%d, 3) = %.3e;\n", prefix, idx, mt.remap[2]);
    PX(fprintf)(comm, file, "%s_fft4(%d, 4)   = %.3e;\n", prefix, idx, mt.fft[3]);
  }
}

static void fprint_average_timer(
    MPI_Comm comm, FILE *restrict file, PX(plan) ths, unsigned flags
    )
{
  if(ths->sign == FFTW_FORWARD)
    fprint_average_timer_internal(comm, file, "pfft_forw", &ths->timer, flags);
  else
    fprint_average_timer_internal(comm, file, "pfft_back", &ths->timer, flags);
}
  
void PX(write_average_timer)(
    MPI_Comm comm, PX(plan) ths, const char *restrict name
    )
{
  int newfile;
  FILE *f;
  
  newfile = !file_exists(name);
  f = open_or_create_file_to_append(comm, name);
  if( newfile )
    write_info_header(comm, f);
  
  write_run_specific_infos(comm, f, ths);
  fprint_average_timer(comm, f, ths, PFFT_PRINT_TIMER_BASIC);
  
  fclose(f);
}

void PX(write_average_timer_adv)(
    MPI_Comm comm, PX(plan) ths, const char *restrict name
    )
{
  FILE *f;
  PX(write_average_timer)(comm, ths, name);
  f = open_or_create_file_to_append(comm, name);
  fprint_average_timer(comm, f, ths, PFFT_PRINT_TIMER_ADV);
  fclose(f);
}

void PX(print_average_timer)(
    MPI_Comm comm, PX(plan) ths
    )
{
  write_info_header(comm, stdout);
  write_run_specific_infos(comm, stdout, ths);
  fprint_average_timer(comm, stdout, ths, PFFT_PRINT_TIMER_BASIC);
}

void PX(print_average_timer_adv)(
    MPI_Comm comm, PX(plan) ths
    )
{
  PX(print_average_timer)(comm, ths);
  fprint_average_timer(comm, stdout, ths, PFFT_PRINT_TIMER_ADV);
}
