/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ipfft.h"

static void set_transpose_parameters_and_permute(
    const INT *n, const INT *local_n,
    const INT *block, int first_dims, unsigned fftw_flags,
    INT *N, INT *blk, PX(permutation) *perm);
static void init_transpose_2d(
    const INT *n, const INT *local_n, const INT *block,
    int first_dims, const PX(permutation) *perm,
    INT *remap_size, INT *remap_block);

INT PX(local_size_transpose_2d)(
    const INT *n, const INT *local_n, INT howmany, const INT *block,
    int first_dims, MPI_Comm comm_cart_1d, unsigned fftw_flags,
    PX(permutation) *perm
    )
{
  INT local_N0_start, local_N0, local_N1_start, local_N1;
  INT N[2], blk[2];

  if( (first_dims != 1) && (first_dims != 2))
    return 1;

  /* set all parameters appropriate for FFTW transpose */
  set_transpose_parameters_and_permute(n, local_n, block, first_dims, fftw_flags, N, blk, perm);

  if(N[0]*N[1]==0)
    return 1;

  return XM(local_size_many_transposed)(2, N, howmany, blk[0], blk[1],
      comm_cart_1d, &local_N0, &local_N0_start, &local_N1, &local_N1_start);
}


X(plan) PX(plan_transpose_2d)(
    const INT *n, const INT *local_n,
    INT tupleSize, const INT *block, C *data_in, C *data_out,
    int first_dims, MPI_Comm comm_cart_1d, unsigned fftw_flags,
    PX(permutation) *perm
    )
{
  INT N[2], blk[2];

  if( (first_dims != 1) && (first_dims != 2) )
    return NULL;

  /* set all parameters appropriate for FFTW transpose */
  set_transpose_parameters_and_permute(n, local_n, block, first_dims, fftw_flags, N, blk, perm);

  return XM(plan_many_transpose)(
      N[0], N[1], 2*tupleSize, blk[0], blk[1], (R*) data_in, (R*) data_out, comm_cart_1d, fftw_flags);
}


void PX(permute_transpose_2d)(
    int first_dims, unsigned remap_flags,
    PX(permutation) *perm
    )
{
  if( (first_dims != 1) && (first_dims != 2) )
    return;

  if( PX(flag_active)(remap_flags, FFTW_MPI_TRANSPOSED_IN) ){
    PX(permute_transposed_in_or_out)(first_dims, perm);
    first_dims = 3 - first_dims; /* set first_dims to opposite value */
  }

  PX(permute_std_transpose_2d)(first_dims, perm);

  if( PX(flag_not_active)(remap_flags, FFTW_MPI_TRANSPOSED_OUT) )
    PX(permute_transposed_in_or_out)(first_dims, perm);
}


void PX(permute_transposed_in_or_out)(
    int first_dims,
    PX(permutation) *perm
    )
{
  int p = first_dims;

  PX(permute)(perm->n[p],  perm->n[(p+1)%3],  perm->n[(p+2)%3],  perm->n);
  PX(permute)(perm->np[p], perm->np[(p+1)%3], perm->np[(p+2)%3], perm->np);
}


void PX(permute_std_transpose_2d)(
    int first_dims,
    PX(permutation) *perm
    )
{
  int h;
  
  h = perm->np[first_dims];
  perm->np[first_dims] = perm->np[0];
  perm->np[0] = h;
}


void set_transpose_parameters_and_permute(
    const INT *n, const INT *local_n,
    const INT *block, int first_dims, unsigned fftw_flags,
    INT *N, INT *blk, PX(permutation) *perm
    )
{
  
  if( PX(flag_active)(fftw_flags, FFTW_MPI_TRANSPOSED_IN) ){
    PX(permute_transposed_in_or_out)(first_dims, perm);
    first_dims = 3 - first_dims; /* set first_dims to opposite value */
  }

  /* set all parameters appropriate for FFTW transpose */
  init_transpose_2d(n, local_n, block, first_dims, perm, N, blk);

  PX(permute_std_transpose_2d)(first_dims, perm);
  
  if( PX(flag_not_active)(fftw_flags, FFTW_MPI_TRANSPOSED_OUT) )
    PX(permute_transposed_in_or_out)(first_dims, perm);
}


static void init_transpose_2d(
    const INT *n, const INT *local_n, const INT *block,
    int first_dims, const PX(permutation) *perm,
    INT *remap_size, INT *remap_block
    )
{
  /* set array size and block size for FFTW */
  remap_size[0]  = n[perm->n[0]];
  remap_size[1]  = n[perm->n[first_dims]];
  remap_block[0] = block[perm->n[0]];
  remap_block[1] = block[perm->n[first_dims]];

  /* multiply size of grouped dims */
  remap_size[2-first_dims]  *= local_n[perm->n[3-first_dims]];
  remap_block[2-first_dims] *= local_n[perm->n[3-first_dims]];
}

